#include <stdio.h>
extern int errno;
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <linux/sockios.h>
#include "divert.h"
#include <time.h>


#define	MAXOPERS	32
#define	REQ_DIVERT_VERSION	0.44
#define	min(a, b)	((a)<(b)?(a):(b))

#define	DCA	divert_cf_arg

#ifdef DEBUG
	#define PARSEDEBUG(a, b)	disp_args(a, b)
#else
	#define PARSEDEBUG(a, b)
#endif	/* DEBUG */


char				version[32]="0.22";
char				dv_version[32];
float				fdv_version;
int					dev_idx=-1;
int					proto=-1;
struct divert_cf	div_cf[MAXOPERS];
struct divert_blk	div_status;
int					nbopers=0;

divert_cf_arg		null_dca;

int display_version();
void get_dv_version();

int parse_device(char **);
int parse_proto(char **);
int parse_ports(char **, int *, int *, int *, int *);
int parse_tcp(char **);
int parse_udp(char **);
int parse_port(char **);
int parse_ip(char **);
int parse_icmp(char **);

void print_usage(char **);


void get_dv_version()
{
	struct divert_cf	div;
	int					fd;
	
	if ((fd=socket(AF_INET, SOCK_STREAM, 0))==-1)
	{
		perror("get_dv_version(): socket()");
		return;
	}
	memset(&div, 0, sizeof(struct divert_cf));
	div.cmd=DIVCMD_GETVERSION;
	div.arg1.ptr=(void *)dv_version;
	if (ioctl(fd, SIOCGIFDIVERT, &div))
	{
		perror("get_dv_version(): ioctl()");
		strcpy(dv_version, "not available");
		fdv_version=-1.0;
		return;
	}
	sscanf(dv_version, "%f", &fdv_version);
	if (fdv_version<(float)REQ_DIVERT_VERSION)
	{
		fprintf(stderr, "\nWARNING: found Frame Diverter %2.2f; I need at least %2.2f !\n",
						fdv_version, REQ_DIVERT_VERSION);
		fprintf(stderr, "         Some operations may not work.\n\n");
	}
	close(fd);
	return;
}

void disp_args(char *func_name, char **argv)
{
	fprintf(stdout, "-->%s(", func_name);
	while(*argv!=NULL)
	{
		fprintf(stdout, "%s", *argv);
		argv++;
		if (*argv!=NULL)
			fprintf(stdout, " ");
	}
	fprintf(stdout, ")\n");
}


char *strlwr(char *arg)
{
	char	*s=arg;

	if (s==NULL)
		return NULL;

	while(*s!='\0')
		*(s++)=tolower(*s);
	return arg;
}

/*
 *
 *
 */
int add_oper(int cmd, divert_cf_arg arg1, divert_cf_arg arg2, divert_cf_arg arg3)
{
	if (nbopers==MAXOPERS)
	{
		fprintf(stderr, "too many actions specified (max: %d)\n", MAXOPERS);
		return -1;
	}
	div_cf[nbopers].cmd=cmd;
	div_cf[nbopers].arg1=arg1;
	div_cf[nbopers].arg2=arg2;
	div_cf[nbopers].arg3=arg3;
	nbopers++;
#ifdef DEBUG
	fprintf(stdout, "DIVCMD_");
	switch(cmd)
	{
		case DIVCMD_DIVERT:		fprintf(stdout, "DIVERT"); break;
		case DIVCMD_IP: 		fprintf(stdout, "IP"); break;
		case DIVCMD_ICMP:		fprintf(stdout, "ICMP"); break;
		case DIVCMD_UDP:		fprintf(stdout, "UDP"); break;
		case DIVCMD_TCP:		fprintf(stdout, "TCP"); break;
		case DIVCMD_TCPDST:		fprintf(stdout, "TCPDST"); break;
		case DIVCMD_TCPSRC:		fprintf(stdout, "TCPSRC"); break;
		case DIVCMD_UDPDST:		fprintf(stdout, "UDPDST"); break;
		case DIVCMD_UDPSRC:		fprintf(stdout, "UDPSRC"); break;
		case DIVCMD_RESET:		fprintf(stdout, "RESET"); break;
		case DIVCMD_GETSTATUS:	fprintf(stdout, "GETSTATUS"); break;
		default: fprintf(stderr, "Uhh ???");
	}
	fprintf(stdout, "(%d, %d, %d) on eth%d\n", arg1, arg2, arg3, dev_idx);
#endif
	return 0;
}


int display_status()
{
	struct divert_cf	dc;
	struct divert_blk	de;
	int					fd;
	int					pcount;
	int					i;

	memset(&dc, 0, sizeof(struct divert_cf));
	memset(&de, 0, sizeof(struct divert_blk));
	
	dc.cmd=DIVCMD_GETSTATUS;
	dc.arg1.ptr=&de;
	dc.dev_index=dev_idx;
	
	if ((fd=socket(AF_INET, SOCK_STREAM, 0))==-1)
	{
		perror("display_status(): socket()");
		return -1;
	}

	if (ioctl(fd, SIOCGIFDIVERT, &dc))
	{
		perror("display_status(): ioctl()");
		return -1;
	}

	fprintf(stdout, "version: %s\n", dv_version);
	fprintf(stdout, "status:  %s\n", de.divert?"active":"inactive");
	fprintf(stdout, "ip:      %s\n", de.protos&DIVERT_PROTO_IP?"yes":"no");
	fprintf(stdout, "icmp:    %s\n", de.protos&DIVERT_PROTO_ICMP?"yes":"no");
	fprintf(stdout, "tcp:     ");
	if (de.protos&DIVERT_PROTO_TCP) 
		fprintf(stdout, "yes\n");
	else
	{
		pcount=0;
		for (i=0; i<MAX_DIVERT_PORTS; i++)
			if (de.tcp_dst[i]>0)
			{
				pcount++;
				fprintf(stdout, "\n         * -> %d", ntohs(de.tcp_dst[i]));
			}
		for (i=0; i<MAX_DIVERT_PORTS; i++)
			if (de.tcp_src[i]>0)
			{
				pcount++;
				fprintf(stdout, "\n         %d -> *", ntohs(de.tcp_src[i]));
			}
		if (pcount==0)
			fprintf(stdout, "no\n");
		else
			fprintf(stdout, "\n");
	}
	fprintf(stdout, "udp:     ");
	if (de.protos&DIVERT_PROTO_UDP)
		fprintf(stdout, "yes\n");
	else
	{
		pcount=0;
		for (i=0; i<MAX_DIVERT_PORTS; i++)
			if (de.udp_dst[i]>0)
			{
				pcount++;
				fprintf(stdout, "\n         * -> %d", ntohs(de.udp_dst[i]));
			}
		for (i=0; i<MAX_DIVERT_PORTS; i++)
			if (de.udp_src[i]>0)
			{
				pcount++;
				fprintf(stdout, "\n         %d -> *", ntohs(de.udp_src[i]));
			}
		if (pcount==0)
			fprintf(stdout, "no\n");
		else
			fprintf(stdout, "\n");
	}
	
	close(fd);
	return 0;
}


/*
 * (dev)ice <eth0> enable|disable||reset|status
 *
 */
int parse_device(char **argv)
{	
	char			devname[32];
	int				argcount=0, ret;
	divert_cf_arg	arg1, arg2, arg3;

	PARSEDEBUG("parse_device", argv);
	
	if (*argv==NULL)
	{
		fprintf(stderr, "device: missing argument !\n");
		return -1;
	}
	
	if (strncmp(*argv, "eth", 3))
	{
		fprintf(stderr, "%s: wrong interface type, expected something like eth??.\n", *argv);
		return -1;
	}
	else
		strcpy(devname, *argv);
	argv++;
	argcount++;

	dev_idx=atoi(&devname[3]);

	/* assume status */
	if (*argv==NULL)
	{
		exit(display_status(0));
	}
	if (!strcmp(*argv, "enable"))
	{
		argcount++;
		argv++;
		arg1.int32=DIVARG1_ENABLE;
		if (add_oper(DIVCMD_DIVERT, arg1, null_dca, null_dca))
			return -1;
	}
	else if (!strcmp(*argv, "disable"))
	{
		argcount++;
		argv++;
		arg1.int32=DIVARG1_DISABLE;
		if (add_oper(DIVCMD_DIVERT, arg1, null_dca, null_dca))
			return -1;
	}
	else if (!strcmp(*argv, "reset"))
	{
		argv++;
		argcount++;
		if (*argv!=NULL)
		{
			fprintf(stderr, "%s: what is this supposed to mean ?\n", *argv);
			return -1;
		}
		if (add_oper(DIVCMD_RESET, null_dca, null_dca, null_dca))
			return -1;
	}
	
	if ((ret=parse_proto(argv))==-1)
		return -1;

	return argcount+ret;
}

/*
 * ip|icmp|tcp|udp
 * 
 */
int parse_proto(char **argv)
{
	int	argcount=0;
	int	oper;
	int	ret;

	PARSEDEBUG("parse_proto", argv);
	
	if (*argv==NULL)
	{
		return 0;
	}
	if (!strcmp(*argv, "tcp"))
	{
		argcount++;
		argv++;
		if ((ret=parse_tcp(argv))==-1)
			return -1;
		argcount+=ret;
	}
	else if (!strcmp(*argv, "udp"))
	{
		argcount++;
		argv++;
		if ((ret=parse_udp(argv))==-1)
			return -1;
		argcount+=ret;
	}
	else if (!strcmp(*argv ,"icmp"))
	{
		argcount++;
		argv++;
		if ((ret=parse_icmp(argv))==-1)
			return -1;
		argcount+=ret;
	}
	else if (!strcmp(*argv, "ip"))
	{
		argcount++;
		argv++;
		if ((ret=parse_ip(argv))==-1)
			return -1;
		argcount+=ret;
	}
	else
	{
		fprintf(stderr, "%s: unknown or unsupported protocol.\n", *argv);
		return -1;
	}
	return argcount;
}


int parse_ports(char **argv, int *dports, int *dport_count, 
				int *sports, int *sport_count)
{
	int		argcount=0, ret, *ports=NULL, *pcount, port;
	char	nullchar='\0', *s=&nullchar;
	
	PARSEDEBUG("parse_ports", argv);

	if (*argv==NULL)
		return 0;
	
	if (*dport_count==MAX_DIVERT_PORTS)
	{
		fprintf(stderr, "too many destination ports specified (max is %d)\n", 
						MAX_DIVERT_PORTS);
		return -1;
	}
	if (*sport_count==MAX_DIVERT_PORTS)
	{
		fprintf(stderr, "too many source ports specified (max is %d)\n", 
						MAX_DIVERT_PORTS);
		return -1;
	}
	
	if (!strcmp(*argv, "dst"))
	{
		ports=dports;
		pcount=dport_count;
	}
	else if (!strcmp(*argv, "src"))
	{
		ports=sports;
		pcount=sport_count;
	}
	else
	{
		if ((ret=parse_proto(argv))==-1)
			return -1;
		else
			return ret;
	}
	argv++;
	argcount++;
	while (*s=='\0' && *argv!=NULL)
	{
		port=(int)strtol(*argv, &s, 10);
		if (*s=='\0')
		{
			if (*pcount==MAX_DIVERT_PORTS)
			{
				fprintf(stderr, "too many ports specified\n");
				return -1;
			}
			ports[*pcount]=port;
			(*pcount)++;
			argv++;
			argcount++;
		}
	}

	if ((ret=parse_ports(argv, dports, dport_count, sports, sport_count))==-1)
		return -1;
	
	return ret+argcount;
}

/*
 * tcp add|del [dst|src <port1> [port2 [port3 [...]]]]
 *
 */
int parse_tcp(char **argv)
{
	int				argcount=0, ret, i;
	int				oper=-1;
	int				dports[MAX_DIVERT_PORTS], dport_count=0,
					sports[MAX_DIVERT_PORTS], sport_count=0;
	divert_cf_arg	arg1, arg2, arg3;

	PARSEDEBUG("parse_tcp", argv);
	
	if (*argv==NULL)
	{
		fprintf(stderr, "tcp: missing argument\n");
		return -1;
	}

	if (!strcmp(*argv, "add"))
		oper=DIVARG1_ADD;
	else if (!strcmp(*argv, "del"))
		oper=DIVARG1_REMOVE;
	else
	{
		fprintf(stderr, "%s: what is this supposed to mean ?\n", *argv);
		return -1;
	}
	argcount++;
	argv++;

	ret=parse_ports(argv, dports, &dport_count, sports, &sport_count);

	switch(ret)
	{
		case 0:
			if (oper==DIVARG1_ADD) oper=DIVARG1_ENABLE;
			if (oper==DIVARG1_REMOVE) oper=DIVARG1_DISABLE;
			arg1.int32=oper;
			
			if (add_oper(DIVCMD_TCP, arg1, null_dca, null_dca))
				return -1;
			break;

		case -1:
			return -1;

		default:
			for (i=0; i<dport_count; i++)
			{
				arg1.int32=oper;
				arg2.uint16=dports[i];
				if (add_oper(DIVCMD_TCPDST, arg1, arg2, null_dca))
					return -1;
			}
			for (i=0; i<sport_count; i++)
			{
				arg1.int32=oper;
				arg2.uint16=sports[i];
				if (add_oper(DIVCMD_TCPSRC, arg1, arg2, null_dca))
					return -1;
			}
			argcount+=ret;
			argv+=ret;
	}
	
	if ((ret=parse_proto(argv))==-1)
		return -1;
	
	return argcount+ret;


}

/*
 * udp [dst|src <port1> [port2 [port3 [...]]]]
 *
 */
int parse_udp(char **argv)
{
	int 			argcount=0, ret, i;
	int				oper=-1;
	int				dports[MAX_DIVERT_PORTS], dport_count=0,
					sports[MAX_DIVERT_PORTS], sport_count=0;
	divert_cf_arg	arg1, arg2, arg3;

	PARSEDEBUG("parse_udp", argv);
	
	if (*argv==NULL)
	{
		fprintf(stderr, "udp: missing argument\n");
		return -1;
	}

	if (!strcmp(*argv, "add"))
		oper=DIVARG1_ADD;
	else if (!strcmp(*argv, "del"))
		oper=DIVARG1_REMOVE;
	else
	{
		fprintf(stderr, "%s: what is this supposed to mean ?\n", *argv);
		return -1;
	}
	argcount++;
	argv++;

	ret=parse_ports(argv, dports, &dport_count, sports, &sport_count);

	switch(ret)
	{
		case 0:
			if (oper==DIVARG1_ADD) oper=DIVARG1_ENABLE;
			if (oper==DIVARG1_REMOVE) oper=DIVARG1_DISABLE;
			arg1.int32=oper;
			if (add_oper(DIVCMD_UDP, arg1, null_dca, null_dca))
				return -1;
			break;

		case -1:
			return -1;

		default:
			for (i=0; i<dport_count; i++)
			{
				arg1.int32=oper;
				arg2.uint16=dports[i];
				if (add_oper(DIVCMD_UDPDST, arg1, arg2, null_dca))
					return -1;
			}
			for (i=0; i<sport_count; i++)
			{
				arg1.int32=oper;
				arg2.uint16=sports[i];
				if (add_oper(DIVCMD_UDPSRC, arg1, arg2, null_dca))
					return -1;
			}
			argcount+=ret;
			argv+=ret;
	}
	
	if ((ret=parse_proto(argv))==-1)
		return -1;
	
	return argcount+ret;
}

int parse_ip(char **argv)
{
	int	argcount=0, ret, oper=-1;
	divert_cf_arg	arg1, arg2, arg3;
	
	PARSEDEBUG("parse_ip", argv);
	
	if (*argv==NULL)
	{
		fprintf(stderr, "ip: missing argument\n");
		return -1;
	}
	
	if (!strcmp(*argv, "add"))
		oper=DIVARG1_ENABLE;
	else if (!strcmp(*argv, "del"))
		oper=DIVARG1_DISABLE;
	else
	{
		fprintf(stderr, "%s: what is this supposed to mean ?\n", *argv);
		return -1;
	}

	argcount++;
	arg1.int32=oper;
	if (add_oper(DIVCMD_IP, arg1, null_dca, null_dca))
		return -1;
	
	if ((ret=parse_proto(argv+argcount))==-1)
		return -1;

	return argcount+ret;
}


int parse_icmp(char **argv)
{
	int	argcount=0, ret, oper=-1;
	divert_cf_arg	arg1, arg2, arg3;
		
	PARSEDEBUG("parse_icmp", argv);
	
	if (*argv==NULL)
	{
		fprintf(stderr, "icmp: missing argument\n");
		return -1;
	}
	if (!strcmp(*argv, "add"))
		oper=DIVARG1_ENABLE;
	else if (!strcmp(*argv, "del"))
		oper=DIVARG1_DISABLE;
	else
	{
		fprintf(stderr, "%s: what is this supposed to mean ?\n", *argv);
		return -1;
	}
	argcount++;
	arg1.int32=oper;
	if (add_oper(DIVCMD_ICMP, arg1, null_dca, null_dca))
		return -1;
	
	if ((ret=parse_proto(argv+argcount))==-1)
		return -1;

	return argcount+ret;
}


void print_usage(char **argv)
{
	time_t	t=(time_t)__DATE__;
	fprintf(stdout, "divert-utils %s, compiled on %s\n", version, t);
	fprintf(stdout, "Usage: %s on <dev> [enable|disable|reset|version] [{PROTO}]>\n", argv[0]);
	fprintf(stdout, "    {PROTO} = <ip add|del> | <icmp add|del> |\n");
	fprintf(stdout, "              <tcp add|del [{PORTS}]> | <udp add|del [{PORTS}]> [{PROTO}]\n");
	fprintf(stdout, "    {PORTS} = <dst|src> <n1 [n2 ... nn] [{PORTS}]>\n");
}

int main(int argc, char **argv)
{
	int		fd, i, err;
	struct parser_block	*pb;

	memset(&null_dca, 0, sizeof(divert_cf_arg));
	
	get_dv_version();
	
	if (argc==1)
	{
		print_usage(argv);
		return -1;
	}

	for (i=1; i<argc; i++)
		strlwr(argv[i]);
	
	argv++;
	
	if (strcmp(*argv, "on"))
	{
		fprintf(stderr, "wrong syntax.\n");
		return -1;
	}

	if (parse_device(argv+1)==-1)
		return -1;

	if ((fd=socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
		perror("socket");
		return -1;
	}
	if (nbopers==0)
	{
		fprintf(stderr, "You didn't specify any action !\n");
		print_usage(argv);
		return -1;
	}
	
	if (dev_idx==-1)
	{
		fprintf(stderr, "You didn't specify a device !\n");
		print_usage(argv);
		return -1;
	}
#ifdef DEBUG
	fprintf(stdout, "performing %d operation(s)\n", nbopers);
#endif
	
	/* Check for minimum version of kernel diverter code */
	
	for (i=0; i<nbopers; i++)
	{
		div_cf[i].dev_index=dev_idx;
		if ((err=ioctl(fd, SIOCSIFDIVERT, (void *)&div_cf[i])))
		{
#ifdef DEBUG
			fprintf(stdout, "Oper[%d]:", i);
#endif
			perror("main(): ioctl()");
			return -1;
		}
	}
	
	close(fd);
	return 0;
}
